// JUMA-PA100 service & setup routines
// Juha Niinikoski, OH2NLT, 08.07.2008

// Display clean up 11.11.2008
// Factory setup function corrected 16.12.2008

#include <stdio.h>					// Added - 5B4AIY
#include <math.h>
#include "juma-pa100.h"
#include "pa100_eeprom.h"			// Get EEPROM structure definitions

// External References
// External Functions
extern void beep(int, int);						// Generate beep, period, length
extern void ms_delay(unsigned int);				// Delay
extern int encoder_get(void);					// Read encoder
extern void clear_lcd(void);					// Clear LCD
extern void lcd_cmd(unsigned char);
extern void lcd_putst(register const char *);	// LCD put string
extern void disp_meter(double, int);			// Display meters - new version 27/JAN/2014
extern void disp_fwd_pwr(void);					// Display wattmeter
extern void disp_id(void);						// Display RF Amp total current
extern void check_alarms(void);					// Check alarms
extern void set_relays(void);					// Set gain & filter relays
extern void analog_measurements(void);			// Do analog measurements
extern void fan_control(void);					// Cooling fan control
extern int convert_adc12(unsigned int);			// Added - 5B4AIY
extern int save_settings(int, int);				// Added - 5B4AIY
extern void set_value(int, int *, int, int);	// Added - 5B4AIY
extern void eval_band(int);						// Frequency Sense function
extern double get_freq(void);					// Get the input frequency
extern void display_beeps(int);					// Page Increment/Decrement beep tone

// External Data
extern long fwd_pwr;				// Relative forward power
extern long rev_pwr;				// Relative reverse power
extern long out_pwr;				// Scaled output power = (fwd_pwr - rev_pwr) * scale factor
extern int pa_state;				// PA Standby/Operate State
extern int id_cur;					// PA current raw value
extern int key;						// Copy of TX request input
extern int svc_flag;
extern unsigned int alarms;			// Alarm bits
extern unsigned int button_timer;	// Timer for buttons
extern char lcdpbuff[];				// General purpose buffer for displays, etc
extern const char *on_off[];
extern const char *auto_man[];
extern const char *bs_txt[];

// Local Data
const char trip_fmt[] = {"Trip:%11s"};
const char trip_voltage[] = {"Limit:%9.2fV"};
const char cal_factor_msg[] = {"Cal Factor:%5i"};
const char cal_offset_msg[] = {"Cal Offset:%5i"};
const char freq_cal_fmt[] = {"Freq: %6.3f MHz"};
const char *cal_prompt[] = {					// Indexed by: cal_page
							"Supply:  ",		// 0  - Voltmeter
							"",					// 1  - Ammeter (Place holder)
							"",					// 2  - RF Power (High Power - Place holder)
							"",					// 3  - RF Power (Low Power - Place holder)
							"Beep Len 0 = Off",	// 4  - Beep Length
							"Power Averaging ",	// 5  - RF Power Measurement Averaging
							"Overvoltage     ",	// 6  - Overvoltage Trip On/Off
							"Overvoltage Trip",	// 7  - Overvoltage Trip Setting
							"Low Voltage     ",	// 8  - Undervoltage Trip On/Off
							"Low Voltage Trip",	// 9  - Undervoltage Trip Setting
							"Full-Scale Power",	// 10 - Graphic Power Meter Full-Scale Setting
							""					// 11 - Frequency Meter calibration factor (place holder)
							};

// Standard Prompts & Messages
const char svc_4_Msg1[] = {"Beep:%9imS"};
const char svc_4_Msg2[] = {"Beep:%11s"};
const char svc_5_Msg[] = {"Samples:%8i"};
const char svc_10_Msg[] = {"Max Power:%5iW"};
const char svc_11_msg[] = {"Factor:%9ld"};
const char Service_Calibration[] = {"\n\rService & Calibration Mode"};
const char Display_Config[] = {"\n\rDISPLAY/CONFIG: Select Page\n\rOPER: Save & Exit\n\r"};
const char Calibration_Saved[] = {"\n\rCalibration settings saved to EEPROM\n\r"};

int cal_page = 0;			// Starting Calibration Page

// User Configuration Values
extern union
	{
	struct defval defval;
	} eeprom;

// System Calibration Values
extern union
	{
	struct calval calval;
	} cal;

void set_alarm_flag(int on, int off)
	{
	int i;

	i = encoder_get();

	if(i > 0) cal.calval.alarm_flags |= on;

	if(i < 0) cal.calval.alarm_flags &= off;
	}

int set_range(long value, long cal_factor, unsigned long max, unsigned long min)
	{
	value *= cal_factor;

	if(value > max) value = max;
	if(value < min) value = min;

	return (int)(value / cal_factor);
	}

void svc_1_svc_2_prompt(void)
	{
	lcd_putst(auto_man[eeprom.defval.auto_band]);
	lcd_putst(bs_txt[eeprom.defval.band]);
	}

void svc_0(void)			// Voltmeter Calibration
	{
	set_value(1, &cal.calval.batt_mult, BATT_MULT_MAX, BATT_MULT_MIN);
	disp_meter(((double)convert_adc12(BATT_CH) * (double)cal.calval.batt_mult), VOLTS);
	sprintf(lcdpbuff, cal_factor_msg, cal.calval.batt_mult);	// Show scale multiplier factor
	}

void svc_1(void)			// Ammeter Calibration
	{
	set_value(1, &cal.calval.id_mult, ID_MULT_MAX, ID_MULT_MIN);
	svc_1_svc_2_prompt();
	disp_id();
	sprintf(lcdpbuff, cal_factor_msg, cal.calval.id_mult);		// Show scale multiplier factor
	}

void svc_2(void)			// RF Power Meter Calibration (High Power)
	{
	set_value(1, &cal.calval.fwd_pwr_mult, HI_PWR_MULT_MAX, HI_PWR_MULT_MIN);
	svc_1_svc_2_prompt();
	disp_fwd_pwr();
	sprintf(lcdpbuff, cal_factor_msg, cal.calval.fwd_pwr_mult);	// Show scale multiplier factor
	}

void svc_3(void)			// RF Power Meter Calibration (Low Power)
	{
	set_value(1, &cal.calval.lo_pwr_offset, LO_PWR_OFFSET_MAX, LO_PWR_OFFSET_MIN);
	svc_1_svc_2_prompt();
	disp_fwd_pwr();
	sprintf(lcdpbuff, cal_offset_msg, cal.calval.lo_pwr_offset);		// Show scale offset factor
	}

void svc_4(void)			// Beep Length Setting
	{
	set_value(1, &cal.calval.beep_len, 100, 0);

	if(cal.calval.beep_len)
		sprintf(lcdpbuff, svc_4_Msg1, cal.calval.beep_len);		// Show current setup
	else
		sprintf(lcdpbuff, svc_4_Msg2, on_off[OFF]);
	}

void svc_5(void)			// Power Averaging Samples
	{
	set_value(1, &cal.calval.samples, SAMPLE_MAX, SAMPLE_MIN);
	sprintf(lcdpbuff, svc_5_Msg, cal.calval.samples);			// Improved display - 5B4AIY
	}

void svc_6(void)			// Overvoltage Trip On/Off
	{
	set_alarm_flag(0x01, 0x02);
	sprintf(lcdpbuff, trip_fmt, (cal.calval.alarm_flags & 0x01) ? on_off[ON] : on_off[OFF]);
	}

void svc_7(void)			// Overvoltage Trip Setting
	{
	cal.calval.overvoltage_trip += encoder_get();
	cal.calval.overvoltage_trip = set_range((long)cal.calval.overvoltage_trip, (long)cal.calval.batt_mult, (OVERVOLTAGE_UPPER * 10000UL), (OVERVOLTAGE_LOWER * 10000UL)); 
	sprintf((char *)lcdpbuff, trip_voltage, ((double)cal.calval.overvoltage_trip * (double)cal.calval.batt_mult) / 1000000.0);
	}

void svc_8(void)			// Undervoltage Trip On/Off
	{
	set_alarm_flag(0x02, 0x01);
	sprintf(lcdpbuff, trip_fmt, (cal.calval.alarm_flags & 0x02) ? on_off[ON] : on_off[OFF]);
	}

void svc_9(void)			// Undervoltage Trip Setting
	{
	cal.calval.undervoltage_trip += encoder_get();
	cal.calval.undervoltage_trip = set_range((long)cal.calval.undervoltage_trip, (long)cal.calval.batt_mult, (UNDERVOLTAGE_UPPER * 10000UL), (UNDERVOLTAGE_LOWER * 10000UL));
	sprintf((char *)lcdpbuff, trip_voltage, ((double)cal.calval.undervoltage_trip * (double)cal.calval.batt_mult) / 1000000.0);
	}

void svc_10(void)			// Graphic Power Meter Maximum Scale Setting
	{
	cal.calval.max_power += encoder_get();
	cal.calval.max_power = set_range((long)cal.calval.max_power, 1L, MAX_SCALE_POWER, MIN_SCALE_POWER);
	sprintf((char *)lcdpbuff, svc_10_Msg, cal.calval.max_power);
	}

void svc_11(void)			// Frequency Counter calibration
	{
	sprintf((char *)lcdpbuff, freq_cal_fmt, get_freq());
	lcd_putst(lcdpbuff);
	cal.calval.freq_cal += (long)encoder_get();

	if(cal.calval.freq_cal > (FREQ_CAL + 250L)) cal.calval.freq_cal = FREQ_CAL + 250L;
	if(cal.calval.freq_cal < (FREQ_CAL - 250L)) cal.calval.freq_cal = FREQ_CAL - 250L;

	sprintf((char *)lcdpbuff, svc_11_msg, cal.calval.freq_cal);
	}

void (*set_svc[])() = {			// Indexed by: cal_page
						svc_0,	// 0	Voltmeter Calibration
						svc_1,	// 1	Ammeter Calibration 
						svc_2,	// 2	RF Power Meter Calibration (High Power)
						svc_3,	// 3	RF Power Meter Calibration (Low Power)
						svc_4,	// 4	Beep Length Setting
						svc_5,	// 5	Power Averaging Samples
						svc_6,	// 6	Overvoltage Trip On/Off
						svc_7,	// 7	Overvoltage Trip Setting
						svc_8,	// 8	Undervoltage Trip On/Off
						svc_9,	// 9	Undervoltage Trip Setting
						svc_10,	// 10	Graphic Power Meter Maximum Scale Setting
						svc_11	// 11	Frequenct meter calibration factor
						};

void display_svc_page(void)
	{
	lcd_cmd(LINE1);									// Set cursor on line 1 and,
	lcd_putst(cal_prompt[cal_page]);				// display current calibration page's prompt.
	set_svc[cal_page]();							// Execute current calibration page's code and
	lcd_cmd(LINE2);									// set the cursor on line 2 and, display
	lcd_putst(lcdpbuff);							// the resultant value.
	}

void change_cal_page(int direction)
	{
	if(direction) cal_page--;
	else cal_page++;

	if(cal_page > MAX_SERVICE_PAGES) cal_page = 0;
	if(cal_page < 0) cal_page = MAX_SERVICE_PAGES;
// This complex expression states:
// If the Low Voltage Alarm is not enabled, or if the High Voltage Alarm is not enabled,
// then skip their adjustment page(s).
		if((cal_page == 6 && !(cal.calval.alarm_flags & 0x01)) || (cal_page == 8 && !(cal.calval.alarm_flags & 0x02)))
			{
			if(direction) cal_page--;
			else cal_page++;
			}
	display_beeps(cal_page);
	}

// Service & Calibration Functions
void service(void)
	{
	int i, rf_gain[11];

	clear_lcd();
	eeprom.defval.serial_test = SERIAL_TEST_OFF;	// See note for v1.05r Build 3
	eeprom.defval.rf_power = _WATT;		// Ensure power meter is set to Watts for calibration

	for(i = 0; i < 11; i++)				// Save current gain settings,
		{
		rf_gain[i] = eeprom.defval.rf_gain[i];
		eeprom.defval.rf_gain[i] = 0;	// and set gain to minimum.
		}

	do	{
// TX request
		key = KEY;						// Copy I/O bit to status flag

// Do measurements & check alarms
		analog_measurements();
		check_alarms();

		if(alarms) svc_flag = FALSE;	// Exit service mode
		else
			{

// Auto band select
			if(eeprom.defval.auto_band == AUTO_BAND)	// AUTO Band Select,
				eval_band(0);							// so use the Frequency Sense mode.
			else if(eeprom.defval.band != TEN_METRES)
				eeprom.defval.band = TEN_METRES;		// MANUAL Band Select, then select 10m.

// Set relays
			set_relays();

// Evaluate TX possibility, needed for RF Power & Drain Current calibration
			if(key && (eeprom.defval.band != NOT_KNOWN) && (cal_page == 1 || cal_page == 2))	// TX request
				{
				pa_state = OPERATE;
				TX_ON = ON;								// RF on
				}
			else
				{
				TX_ON = OFF;							// RF off
				pa_state = STANDBY;
				}

// Cooling Fan
			fan_control();
			button_timer = LONG_PUSH;

// Read Switches
			if(!DISP)									// Select calibration page
				{
				while(!DISP)							// SW1
					{
					if(button_timer == 0)
						{
						do	{
							change_cal_page(INCREMENT);
							display_svc_page();
							ms_delay(PAGE_CHANGE);
							} while(!DISP);
						break;
						}
					}

				if(button_timer) change_cal_page(INCREMENT);

				while(!DISP);							// Wait for DISPLAY button to be released
				ms_delay(BUTTON_DEBOUNCE);
				encoder_get();							// Flush encoder
				}

			if(PWR_SW)									// Select calibration page
				{
				while(PWR_SW)							// SW7
					{
					if(button_timer == 0)
						{
						do	{
							change_cal_page(DECREMENT);
							display_svc_page();
							ms_delay(PAGE_CHANGE);
							} while(PWR_SW);
						break;
						}
					}

				if(button_timer) change_cal_page(DECREMENT);

				while(PWR_SW);							// Wait for PWR button to be released
				ms_delay(BUTTON_DEBOUNCE);
				encoder_get();							// Flush encoder
				}

			display_svc_page();							// Now display the service page and its calibration data

			if(!OPER)									// Test SW2 (OPER)
				{
				beep(HZ587_31, cal.calval.beep_len);	// D tone beep for OPER
				i = save_settings(0, 1);

				while(!OPER);							// Wait while OPER button pressed

				cal_page = 0;							// Reset page
				svc_flag = FALSE;						// Exit service mode
				}	// End IF
			}
		} while(svc_flag);

	for(i = 0; i < 11; i++) eeprom.defval.rf_gain[i] = rf_gain[i];	// Restore original gain settings

	pa_state = STANDBY;									// Reset state to STANDBY
	}	// End function

