// dsPIC30 timers PWM DAC for JUMA-PA100
// Juha Niinikoski, OH2NLT 06.07.2008

// Microchip C30 compiler conversion

// JUMA-PA100 Timer & PWM system
/*
 Timer & PWM unit usage
 TMR1 is used for frequency counter
 TMR2 is used for tone generation, IRQ driven system, RD1/OC2 = tone out
 TMR3 is used for 1ms timer tick IRQ and for LCD PWM DAC timebase
 OC3 output is for LCD back light control
 OC4 output is for LCD contrast
 OC3 and OC4 are connected to TMR3
 TMR4 
 TMR5 used for delays (separate module)
 Encoder simulator repeat logic added 13.08.2008
*/
#include "juma-pa100.h"
#include "pa100_eeprom.h"			// Get EEPROM structure definitions

//#define TICK_PERIOD 7500			// 1ms tick
//#define TICK_PERIOD		(FCY / 1000)// 1ms tick (Old setting, prior to frequency counter modification)
#define TICK_PERIOD		7378		// New period, frequency counter

// External Data
extern const unsigned int band_limits[];	// Band limits defined in the main
extern unsigned int alarms;			// Alarm flag
extern int loop_ctr;				// Alarm blink counter
extern int dsp_alarm;				// Alarm message display flag. Set in display_alarms()
extern int pa_state;
extern int beep_flag;
extern long input_freq;

// Local Data
unsigned int freq;					// Frequency counter
unsigned int kx3_timeout;
unsigned int band_query_timer;		// Band data query timer
unsigned int button_timer;			// Long push timer, count from set value to zero, global visibility
int cmd_timeout;					// RS-232 command timeout, 0 = timeout, negative values = state flags.
int decay_counter;					// Power meter slow decay

// Band selector
int band_bins[10];					// Found samples
int freq_sample_ctr;				// Sample counter
int freq_ctr;						// Input frequency sample averaging counter

static unsigned int tone_counter; 	// Beep tone length (ms)
static unsigned int busy_counter;	// Tone generator busy(ms), busy if != 0
static int last_cycle;				// For click-less end trick

// Encoder Simulator
static unsigned int up;				// UP button shift register
static unsigned int dn;				// DOWN button shift register
static int up_pushed;				// Up pushed status
static int dn_pushed;				// Down pushed status
static int rep_up;					// Repeat timers (ms)
static int rep_dn;
static int rep_up_dly;				// Repeat start delay (ms)
static int rep_dn_dly;
int enc = 0;						// Encoder accumulator

// EEPROM structures
extern union
	{
	struct defval defval;
	unsigned int storage[sizeof(struct defval) / 2];
	} eeprom;

// Simulate encoder with UP / DOWN buttons
// Read encoder
int encoder_get(void)
	{
	int temp;

	temp = enc;
	enc = 0;
	return(temp);
	}

// Initialise timer & PWM system
void init_timers_pwm(void)
	{
	freq_sample_ctr = F_SAMPLES;
	freq_ctr = 100;
	OC3CON = 0;						// First turn off the module if it was on
	OC4CON = 0;

// LCD PWM DAC setup
	OC3RS = DEFAULT_BL;				// Set to 300
	OC3R = DEFAULT_BL;				// Set to 300
	OC3CON = 0x000E;				// Use TMR3, OC3 = PWM mode

	OC4RS = DEFAULT_CONTRAST;		// Set to 2000
	OC4R = DEFAULT_CONTRAST;		// Set to 2000
	OC4CON = 0x000E;				// Use TMR3, OC4 = PWM mode

//	PR3 = TICK_PERIOD - 1;			// 7372 Set PWM cycle, TMR3 period register, 1000Hz/1ms
	PR3 = TICK_PERIOD;				// Optimum value for frequency counter. (7378 giving a 1mS timer period = 1000.7053uS)
	T3CONbits.TON = 1;				// Start TMR3
	_T3IE = 1;						// Enable TMR3 tick IRQ

// Timer 1 for frequency counting
	_T1IE = 0;
	T1CON = 0x0002;					// External clock, no sync
	kx3_timeout = 0;
	}

// Tone generator, TMR2
void __attribute__((interrupt)) _T2Interrupt(void)
	{
	switch(last_cycle)				// Tone generator state machine
		{
		case 0: default:			// Tone output off, tri-state
			TONE_TRIS = 1;			// Tri-state output
			_T2IE = 0;				// Disable TMR1 IRQs, Tone generator off
		break;

		case 1:						// Last half cycle
			TONE_OUT = !TONE_OUT;
			PR2 = PR2 >> 1;			// Next = half cycle
			last_cycle = 0;			// Flag for shut down
		break;

		case 2:						// Last full cycle, start shut down
			TONE_OUT = !TONE_OUT;	// Inverse tone out on every cycle

			if(TONE_OUT != 0)		// End with down going half cycle
				last_cycle = 1;		// Next = last full cycle
		break;

		case 3:						// Generate sound
			TONE_OUT = !TONE_OUT;	// Inverse tone out on every cycle
		break;
		}

	_T2IF = 0;
	}

// Tone set functions
void tone_on(int tone)				// Set tone, set value = 7,5MHz / tone(Hz) * 2
	{
	last_cycle = 3;					// Not last cycle, run state
	PR2 = tone - 1;					// Set tone period
	TMR2 = PR2 >> 1;				// Start with half period, reduce click
	TONE_OUT = 0;					// Start allways same phase
	TONE_TRIS = 0;					// Enable output
	T2CONbits.TON = 1;				// Start timer
	_T2IF = 0;						// Clear old pending interrupt, start with clean half cycle
	_T2IE = 1;						// Enable TMR1 IRQs, Tone generator on
	}

void beep(int tone, int duration)
	{
	if(busy_counter == 0 && duration > 0)	// Start tone only if tone generator free & duration specified
		{
		tone_counter = duration;	// Set tone length
		tone_on(tone);				// Set tone pitch & start play
		}
	}
/*
 Frequency Counter
 The frequency counter relies on the 1mS timer interrupt being accurate. This interrupt is generated whenever
 the acmmulated count in the TMR3 period register is equalled. At this point an interrupt is generated, the
 counter cleared and the count of the main clock cycles re-commences. The crystal oscillator used for the main
 clock uses a 7.3728MHz HC6/U crystal, giving a clock period of: 1/7.3728 = 135.6336806nS. In theory, in order
 to generate an interrupt every 1,000uS the period register would need to be set to:

	1000uS / .1356336806uS = 7372

 This would give an actual period of:

	7372 * 135.6336806 = 999.89149uS, or 999891.49nS

 If this is used for the gate time then there will be a constant error, as too few cycles would be counted.
 For example, for an input frequency of 28.85MHz and a gate time of 999891.49nS the accumulated count in the
 frequency counter would be: 28846 +/-1.

 The actual measured counts versus various register settings were:
 
  I/P FREQ			COUNT		COUNT		COUNT		COUNT
	kHz				7272		7373		7378		7379
 ============================================================
	1900			1897		1897		1899		1899
	3600			3596		3596		3599		3599
	7100			7092		7093		7099		7100
	10125			10114		10116		10124		10125
	14200			14185		14187		14199		14201
	18118			18100		18102		18117		18119
	21300			21278		21281		21299		21302
	24940			24915		24918		24939		24942
	28850			28821		28825		28849		28853
 ============================================================
 
 In addition, servicing the interrupt, stopping the count, count transfer, and re-starting the frequency counter
 also consume machine cycles which further shortens the effective gate time. For example, as the table above shows,
 with a timer counter register setting of 7372, a 28.850MHz input is actually counted as 28,821 events. The period
 of this frequency is:

	1 / 28.850 = 0.0346620451nS

 Since the count is 29 cycles short, this means the gate time was 1.0051993068uS short of 1,000uS or an
 effective time of:
 
	28821 * 0.0346620451 = 998994.800694nS

 The actual timer interrup is: 999891.49nS, this means that the actual interrupt servicing, stopping,
 counter transfer and re-starting accounts for:

	999891.493059 - 998994.800694 = 896.6923650nS

 This suggests that between 6 or 7 machine cycles are required. Therefore, to account for this, the timer
 register setting needs to be increased by 6 or 7 to 7378 or 7379, and as the table above shows, the best
 choice is 7378.

 To ensure a stable display, the counts are averaged over 100 samples, and the resultant count then divided
 by 100. However, if this count were simply converted to a floating point number and then displayed with the
 correct decimal point, there is a constant -1 count error because the count represents a truncated count.

 In the Service Module, the calibration constant is nominally 100,000. The 100 portion being the averaging
 factor, and the 1,000 portion being the value required to divide the result to display MHz instead of kHz.

 This calibration factor is adjustable over the limits +/-250 which gives a frequency correction of about
 +/-72kHz at 30MHz and +/-5kHz at 1.9MHz. (See comments in get_freq() in juma-pa100.c)

 Frequency Sense Mode
 If the AUTO band detection is set to the Frequency Sense mode, then the input signal to the amplifier is
 shaped by a pair of diodes, and coupled to the input of the microprocessor. The gate time is 1mS determined
 by the 1mS timer interrupt. On entry to the interrupt the counter is stopped, and the accumulated count is
 transferred to the variable freq. Then the counter is cleared and re-enabled for the next 1mS. Since the gate
 is open for 1mS, the accumulated count represents kHz. For example, for an input signal of 1.8MHz, the counter
 will have accumulated 1,800 counts. The freq_sample_ctr is TRUE so the accumulated count in freq is
 compared with the values already stored in band_limits[]. The freq count is sequentially compared with the
 limit values pre-defined, and if it is less than the current limit, then the condition is met, and the loop
 terminated by a break instruction. The current loop counter index represents the band, and the count in
 the indexed value of the band_bins[] is incremented. The freq_sample counter is decremented and, if it is
 non-zero, the cycle repeats. If it is zero, this represents the end of the measurement cycle. In the main
 module, in the function: eval_band(), if freq_sample_ctr is 0, then we examine the various frequency count
 'bins'. band_bin[0] contains the count of the number of zero count samples. If it is less than 5 then we have
 accumulated enough samples to form a valid band. Each bin is examined from the highest frequency downwards,
 and the first bin to contain a count is considered to be the correct band. The variable eeprom.defval.band
 is assigned the band number, and then the bins are cleared, and freq_sample_ctr is set to the required number
 of counts, 20.
*/
// IRQ code
// 1ms Tick timer
void __attribute__((interrupt)) _T3Interrupt(void)
	{
	int i = 0;

	_T3IF = 0;
//	IRQ_TEST = 1;						// Timing test

// Frequency counter
	T1CONbits.TON = 0;					// Stop counting
	freq = TMR1;						// Save frequency
	TMR1 = 0;							// Reset counter
	T1CONbits.TON = 1;					// Start counter

// Although the 20 samples used by the Band Select module could be used to average the frequency counter readings,
// I found that this gave a somewhat 'glitchy' display. This longer averaging and subsequent display interval leads
// to a much more stable display. A.Ryan - 13/NOV/2014
	if(freq_ctr)
		{
		input_freq += (long)freq;
		freq_ctr--;
		}

// Band select
	if(freq_sample_ctr)					// Run only when previous measurement has been used
		{
		do	{
			if(freq <= band_limits[i])
				{
				band_bins[i]++;			// Sample found, increment the bin count,
				break;					// and exit. (Under normal circumstances, i will never be greater than 9.)
				}						// If the frequency is higher than 30MHz, the break will not be taken, and i will terminate at 10.
			} while (++i < 10);

		if(i == 10)						// Frequency was higher than 30MHz, causing i to increment to 10,
			{
			pa_state = STANDBY;			// so, force state to Standby and set band to NOT_KNOWN. (This should never occur!)
			eeprom.defval.band = NOT_KNOWN;
			}

		freq_sample_ctr--;
		}

// Tone duration counter
	if(tone_counter)
		{
		tone_counter--;

		if(tone_counter == 0) last_cycle = 2;
		}

// Push button timer
	if(button_timer) button_timer--;	// Decrement button timer if active

// Power meter decay counter
	if(decay_counter) decay_counter--;

// Band data query timer
	if(band_query_timer) band_query_timer--;

// Elecraft KX-3 Comms Timeout
// Invoke the comms timer only if polling is disabled
	if(kx3_timeout && !eeprom.defval.poll_timer) kx3_timeout--;

// Alarm blink timer
	if(loop_ctr) loop_ctr--;

	if(loop_ctr == (LOOP_COUNT / 2))
		{
		dsp_alarm = FALSE;				// Display blank line
		beep_flag = TRUE;
		}
/*
 In the PA100 we do not have rotary encoder, therefore the encoder functions are simulated, with UP & DOWN buttons
*/
// Do debounce shift registers
	if(!alarms)							// Ignore buttons if there are any alarms.
		{
		up <<= 1;						// UP button, encoder ++

		if(UP == 0) up |= 0x0001;
		else up &= 0xFFFE;

		dn <<= 1;						// DOWN button, encoder --

		if(DN == 0) dn |= 0x0001;
		else dn &= 0xFFFE;

// Test pushed condition
// UP
		if(up_pushed == 0)				// Not pushed
			{
			if(up == 0xFFFF)			// Test if de-bounce shift register all ones
				{
				up_pushed = 1;			// Set pushed
				rep_up_dly = REP_DLY;	// Set repeat start timer
				enc++;					// Increment encoder
				}
			}
		else							// Pushed
			{
			if(up == 0x0000)			// Test if de-bounce shift register all zeroes
				{
				up_pushed = 0;			// Clear pushed
				rep_up_dly = 0;			// Clear repeat
				rep_up = 0;
				}
			}

// DOWN
		if(dn_pushed == 0)				// Not pushed
			{
			if(dn == 0xFFFF)			// Test if de-bounce shift register all ones
				{
				dn_pushed = 1;			// Set pushed
				rep_dn_dly = REP_DLY;	// Set repeat start timer
				enc--;					// Decrement encoder
				}
			}
		else							// Pushed
			{
			if(dn == 0x0000)			// Test if de-bounce shift register all zeroes
				{
				dn_pushed = 0;			// Clear pushed
				rep_dn_dly = 0;			// Clear repeat
				rep_dn = 0;
				}
			}
// Repeat start timers
		if(rep_up_dly)					// Check if UP button start repeat timer is running
			{
			rep_up_dly--;

			if(rep_up_dly == 0)			// Start repeat action
				rep_up = REP;
			}

		if(rep_dn_dly)					// Check if DOWN button start repeat timer is running
			{
			rep_dn_dly--;

			if(rep_dn_dly == 0)			// Start repeat action
				rep_dn = REP;
			}
// Repeat generator timers
		if(rep_up)						// UP button repeat counter going ?
			{
			rep_up--;

			if(rep_up == 0)
				{
				rep_up = REP;			// Reload timer
				enc++;					// Add encoder accumulator
				}
			}

		if(rep_dn)						// DOWN button repeat counter going ?
			{
			rep_dn--;

			if(rep_dn == 0)
				{
				rep_dn = REP;			// Reload timer
				enc--;					// Subtract encoder accumulator
				}
			}
		}
//	IRQ_TEST = 0;						// Timing test
	}

